/*
 *  Status monitor for Canon NCAP Printer.
 *  Copyright (C) 2004-2013 CANON INC.
 *  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "uimain.h"
#include "widgets.h"
#include "interface.h"
#include "overridedlg.h"

#include "callbacks.h"
#include "support.h"
#include "id_defines.h"
#include "commandcontroller.h"
#include "value_defines.h"

#define OVERRIDE_CTRL_TYPE_NUM 3
#define OVERRIDE_CTRL_TBL_NUM 3

static int InitOverrideDlgWidgets( UIStatusWnd* const wnd );
static unsigned int SetItemValue_OverrideDlgOK( const UIStatusWnd* const wnd );

CtrlTbl* PPDCtrlTbl[CTRL_LIST_MAXNUM];

static CtrlTbl ctrlTbl[OVERRIDE_CTRL_TBL_NUM] =
{
	{ ID_OVERRIDE_DIALOG1, LABEL_TYPE_TITLE, NULL, N_("Paper Size Override Settings"), NULL, 0 },
	{ ID_CTRLOVERRIDEENABLE, LABEL_TYPE_BUTTON, "OverrideDlg_Use_checkbutton", N_("Override Paper Size"), "OverrideDlg_Use_checkbutton", ID1002 },
	{ -1, -1, NULL, NULL, NULL, -1 }
};

UIOverrideDlg* CreateOverrideDlg(UIDialog* const parent)
{
	UIOverrideDlg *pDialog;

	pDialog = (UIOverrideDlg *)CreateDialog(sizeof(UIOverrideDlg), parent);
	if( pDialog != NULL )
	{
		UI_DIALOG(pDialog)->window = create_Override_dialog();
	}
	return pDialog;
}

void ShowOverrideDlg( UIStatusWnd* const wnd )
{
	int nRet = 0;

	if( wnd == NULL )
	{
		return;
	}
	if( wnd->override_dlg == NULL )
	{
		wnd->override_dlg = CreateOverrideDlg( UI_DIALOG( wnd ) );
	}

	SigDisable();
	nRet = InitOverrideDlgWidgets( wnd );
	SigEnable();

	if( nRet == 0 )
	{
		ShowDialog( (UIDialog *)wnd->override_dlg, NULL );
	}
	else
	{
		ShowMsgDlg( wnd, ID_COMMUNICATION_ERR_GET );
		if( wnd->override_dlg != NULL )
		{
			if( wnd->override_dlg->pDialogDict != NULL )
			{
				DeleteDict( wnd->override_dlg->pDialogDict );
				wnd->override_dlg->pDialogDict = NULL;
			}
			mem_free( wnd->override_dlg );
			wnd->override_dlg = NULL;
		}
	}
}

static unsigned int SetItemValue_OverrideDlgOK( const UIStatusWnd* const wnd )
{
	int i = 0;
	int nValue = 0;
	unsigned int unRet = 0;
	GtkWidget *pWindow = NULL;

	if( wnd == NULL )
	{
		UI_DEBUG("SetItemValue_OverrideDlgOK PPDCtrlTbl[%p] pWindow[%p]\n", PPDCtrlTbl, pWindow);
		return DICT_SET_RETURN_ERROR;
	}
	pWindow = wnd->override_dlg->dialog.window;

	for( i = 0; PPDCtrlTbl[i] != NULL; i++ )
	{
		switch( PPDCtrlTbl[i]->dbid )
		{
		case ID1002:
			nValue = GetToggleButtonActive( pWindow, "OverrideDlg_Use_checkbutton", FALSE );
			unRet |= SetItemValuetype_int( wnd->override_dlg->pDialogDict, PPDCtrlTbl[i]->dbid, nValue );
			UI_DEBUG("SetItemValue_OverrideDlgOK unRet[%d]\n", unRet);
			break;
		default:
			break;
		}
		if( (unRet & DICT_SET_RETURN_ERROR) == DICT_SET_RETURN_ERROR )
		{
			break;
		}
	}
	return unRet;
}

void OverrideDlgOK( UIStatusWnd* const wnd )
{
	int nRet = 0;
	unsigned int unRet = 0;

	if( wnd == NULL )
	{
		nRet = -1;
	}
	else
	{
		nRet = CreateDict_SetData( wnd->override_dlg->pDialogDict );
	}

	if( nRet == 0 )
	{
		unRet = SetItemValue_OverrideDlgOK( wnd );
		if( (unRet & DICT_SET_RETURN_ERROR) == DICT_SET_RETURN_ERROR )
		{
			nRet = -1;
		}
	}

	UI_DEBUG("OverrideDlgOK unRet[%d]\n", unRet);
	if( ( nRet == 0 ) && ( unRet == DICT_SET_RETURN_CHANGE ) && (wnd != NULL) )
	{
		nRet = CommunicatePrinterData( wnd, (UIDialog *)wnd->override_dlg, wnd->override_dlg->pDialogDict, CODE_COMMAND_ANALYZE, TRUE );
		UI_DEBUG("OverrideDlgOK CommunicatePrinterData[%d]\n", nRet);
	}

	if( wnd != NULL )
	{
		HideDialog( (UIDialog *)wnd->override_dlg );
	}

	if( nRet != 0 )
	{
		ShowMsgDlg( wnd, ID_COMMUNICATION_ERR_SET );
	}

	return;
}

static int InitOverrideDlgWidgets( UIStatusWnd* const wnd )
{
	GtkWidget *pWindow = NULL;
	GList *pGlist = NULL;
	int nValue = 1;
	int i = 0;
	int nRet = 0;
	const char* const ctrl_type[OVERRIDE_CTRL_TYPE_NUM] = { "CNSUIPORIDEDlg", "CNSUICtrlOverride", NULL };

	if( wnd == NULL )
	{
		return -1;
	}

	if( wnd->override_dlg == NULL )
	{
		return -1;
	}
	else
	{
		pWindow = wnd->override_dlg->dialog.window;
	}

	CreatePPDCtrlTbl( wnd, ctrl_type, ctrlTbl, PPDCtrlTbl );

	SetLabel_HideWidget( pWindow, PPDCtrlTbl, ctrlTbl );

	pGlist = CreateGlist_DbId( PPDCtrlTbl );

	if( wnd->override_dlg->pDialogDict != NULL )
	{
		UI_DEBUG("pDialogDict no mem_free, from Func[%s] Line[%u] \n", __func__, __LINE__);
		DeleteDict( wnd->override_dlg->pDialogDict );
	}
	wnd->override_dlg->pDialogDict = CretateDict_GetData( pGlist );
	g_list_free_wrapper( pGlist );
	pGlist = NULL;

	if( wnd->override_dlg->pDialogDict == NULL )
	{
		return -1;
	}

	nRet = CommunicatePrinterData( wnd, NULL, wnd->override_dlg->pDialogDict, CODE_COMMAND_ANALYZE, TRUE );
	if( nRet != 0 )
	{
		return -1;
	}

	for( i = 0; PPDCtrlTbl[i] != NULL; i++ )
	{
		switch( PPDCtrlTbl[i]->dbid )
		{
		case ID1002:
			GetItemValueType_int( wnd->override_dlg->pDialogDict, PPDCtrlTbl[i]->dbid, &nValue );
			SetActiveCheckButton( pWindow, "OverrideDlg_Use_checkbutton" , nValue, FALSE );
			break;
		default:
			break;
		}
	}
	return nRet;
}

void DisposeOverRideDlg( UIStatusWnd* const wnd )
{
	UIOverrideDlg* ui_override_dlg = NULL;

	if(wnd == NULL)
	{
		return;
	}

	ui_override_dlg = wnd->override_dlg;
	if( ui_override_dlg != NULL )
	{
		if( ui_override_dlg->pDialogDict != NULL )
		{
			DeleteDict( ui_override_dlg->pDialogDict );
			ui_override_dlg->pDialogDict = NULL;
		}
		DisposeDialog( (UIDialog *)ui_override_dlg );
		wnd->override_dlg = NULL;
	}
}
